@@ -24,6 +24,49 @@
  * Contributor(s): ______________________________________.
  *
  * $Log: safecoll.h,v $
+ * Revision 1.15  2005/11/25 03:43:47  csoutheren
+ * Fixed function argument comments to be compatible with Doxygen
+ *
+ * Revision 1.14  2004/11/08 02:34:18  csoutheren
+ * Refactored code to (hopefully) compile on Linux
+ *
+ * Revision 1.13  2004/11/07 12:55:38  rjongbloed
+ * Fixed safe ptr casting so keeps associated collection for use in for loops.
+ *
+ * Revision 1.12  2004/10/28 12:19:44  rjongbloed
+ * Added oeprator! to assure test for NULL that some people use is correct for PSafePtr
+ *
+ * Revision 1.11  2004/10/14 12:31:45  rjongbloed
+ * Added synchronous mode for safe collection RemoveAll() to wait until all objects
+ *   have actually been deleted before returning.
+ *
+ * Revision 1.10  2004/10/04 12:54:33  rjongbloed
+ * Added functions for locking an unlocking to "auto-unlock" classes.
+ *
+ * Revision 1.9  2004/08/12 12:37:40  rjongbloed
+ * Fixed bug recently introduced so removes deleted object from deletion list.
+ * Also changed removal list to be correct type.
+ *
+ * Revision 1.8  2004/08/05 12:15:56  rjongbloed
+ * Added classes for auto unlocking read only and read write mutex on
+ *   PSafeObject - similar to PWaitAndSIgnal.
+ * Utilised mutable keyword for mutex and improved the constness of functions.
+ * Added DisallowDeleteObjects to safe collections so can have a PSafeObject in
+ *   multiple collections.
+ * Added a tempalte function to do casting of PSafePtr to a PSafePtr of a derived
+ *   class.
+ * Assured that a PSafeObject present on a collection always increments its
+ *   reference count so while in collection it is not deleted.
+ *
+ * Revision 1.7  2002/12/10 07:36:57  robertj
+ * Fixed possible deadlock in PSafeCollection find functions.
+ *
+ * Revision 1.6  2002/10/29 00:06:14  robertj
+ * Changed template classes so things like PSafeList actually creates the
+ *   base collection class as well.
+ * Allowed for the PSafeList::Append() to return a locked pointer to the
+ *   object just appended.
+ *
  * Revision 1.5  2002/10/04 08:22:40  robertj
  * Changed read/write mutex so can be called by same thread without deadlock
  *   removing the need to a lock count in safe pointer.
@@ -233,17 +276,55 @@ class PSafeObject : public PObject
        This is typically used by the PSafeCollection class and is not expected
        to be used directly by an application.
       */
-    BOOL SafelyCanBeDeleted();
+    BOOL SafelyCanBeDeleted() const;
   //@}
 
   protected:
-    PMutex          safetyMutex;
-    unsigned        safeReferenceCount;
-    BOOL            safelyBeingRemoved;
-    PReadWriteMutex safeInUseFlag;
+    mutable PMutex          safetyMutex;
+            unsigned        safeReferenceCount;
+            BOOL            safelyBeingRemoved;
+    mutable PReadWriteMutex safeInUseFlag;
 };
 
 
+/**Lock a PSafeObject for read only and automatically unlock it when go out of scope.
+  */
+class PSafeLockReadOnly
+{
+  public:
+    PSafeLockReadOnly(const PSafeObject & object);
+    ~PSafeLockReadOnly();
+    BOOL Lock();
+    void Unlock();
+    BOOL IsLocked() const { return locked; }
+    bool operator!() const { return !locked; }
+
+  protected:
+    PSafeObject & safeObject;
+    BOOL          locked;
+};
+
+
+
+/**Lock a PSafeObject for read/write and automatically unlock it when go out of scope.
+  */
+class PSafeLockReadWrite
+{
+  public:
+    PSafeLockReadWrite(const PSafeObject & object);
+    ~PSafeLockReadWrite();
+    BOOL Lock();
+    void Unlock();
+    BOOL IsLocked() const { return locked; }
+    bool operator!() const { return !locked; }
+
+  protected:
+    PSafeObject & safeObject;
+    BOOL          locked;
+};
+
+
+
 /** This class defines a thread-safe collection of objects.
   This class is a wrapper around a standard PCollection class which allows
   only safe, mutexed, access to the collection.
@@ -266,7 +347,7 @@ class PSafeCollection : public PObject
        Note the collection is automatically deleted on destruction.
      */
     PSafeCollection(
-      PCollection * collection    /// Actual collection of objects
+      PCollection * collection    ///< Actual collection of objects
      );
 
     /**Destroy the thread safe collection.
@@ -278,14 +359,6 @@ class PSafeCollection : public PObject
   /**@name Operations */
   //@{
   protected:
-    /**Add an object to the collection.
-       This uses the PCollection::Append() function to add the object to the
-       collection, with full mutual exclusion locking on the collection.
-      */
-    virtual PINDEX SafeAppend(
-      PSafeObject * obj     /// Object to add to collection
-    );
-
     /**Remove an object to the collection.
        This function removes the object from the collection itself, but does
        not actually delete the object. It simply moves the object to a list
@@ -295,7 +368,7 @@ class PSafeCollection : public PObject
        is maintained.
       */
     virtual BOOL SafeRemove(
-      PSafeObject * obj   /// Object to remove from collection
+      PSafeObject * obj   ///< Object to remove from collection
     );
 
     /**Remove an object to the collection.
@@ -306,40 +379,72 @@ class PSafeCollection : public PObject
        As for Append() full mutual exclusion locking on the collection itself
        is maintained.
       */
-    virtual PSafeObject * SafeRemoveAt(
-      PINDEX idx    /// Object index to remove
+    virtual BOOL SafeRemoveAt(
+      PINDEX idx    ///< Object index to remove
     );
 
   public:
     /**Remove all objects in collection.
       */
-    virtual void RemoveAll();
+    virtual void RemoveAll(
+      BOOL synchronous = FALSE  ///< Wait till objects are deleted before returning
+    );
+
+    /**Disallow the automatic delete any objects that have been removed.
+       Objects are simply removed from the collection and not marked for
+       deletion using PSafeObject::SafeRemove() and DeleteObject().
+      */
+    void AllowDeleteObjects(
+      BOOL yes = TRUE   ///< New value for flag for deleting objects
+    ) { deleteObjects = yes; }
+
+    /**Disallow the automatic delete any objects that have been removed.
+       Objects are simply removed from the collection and not marked for
+       deletion using PSafeObject::SafeRemove() and DeleteObject().
+      */
+    void DisallowDeleteObjects() { deleteObjects = FALSE; }
 
     /**Delete any objects that have been removed.
+       Returns TRUE if all objects in the collection have been removed and
+       their pending deletions carried out.
       */
-    virtual void DeleteObjectsToBeRemoved();
+    virtual BOOL DeleteObjectsToBeRemoved();
+
+    /**Delete an objects that has been removed.
+      */
+    virtual void DeleteObject(PObject * object) const;
 
     /**Start a timer to automatically call DeleteObjectsToBeRemoved().
       */
     virtual void SetAutoDeleteObjects();
 
     /**Get the current size of the collection.
-       Note that use of this function is limited as it is not subject to the
-       collections mutual exclusion locking. Thus, it is merely an
+       Note that usefulness of this function is limited as it is merely an
        instantaneous snapshot of the state of the collection.
       */
-    PINDEX GetSize() const { return collection->GetSize(); }
+    PINDEX GetSize() const;
+
+    /**Determine if the collection is empty.
+       Note that usefulness of this function is limited as it is merely an
+       instantaneous snapshot of the state of the collection.
+      */
+    BOOL IsEmpty() const { return GetSize() == 0; }
+
+    /**Get the mutex for the collection.
+      */
+    const PMutex & GetMutex() const { return collectionMutex; }
   //@}
 
   protected:
     void SafeRemoveObject(PSafeObject * obj);
     PDECLARE_NOTIFIER(PTimer, PSafeCollection, DeleteObjectsTimeout);
 
-    PCollection * collection;
-    PMutex        collectionMutex;
-    PAbstractList toBeRemoved;
-    PMutex        removalMutex;
-    PTimer        deleteObjectsTimer;
+    PCollection  *     collection;
+    mutable PMutex     collectionMutex;
+    BOOL               deleteObjects;
+    PList<PSafeObject> toBeRemoved;
+    PMutex             removalMutex;
+    PTimer             deleteObjectsTimer;
 
   friend class PSafePtrBase;
 };
@@ -381,8 +486,8 @@ class PSafePtrBase : public PObject
        and -- operators will not work.
      */
     PSafePtrBase(
-      PSafeObject * obj = NULL,         /// Physical object to point to.
-      PSafetyMode mode = PSafeReference /// Locking mode for the object
+      PSafeObject * obj = NULL,         ///< Physical object to point to.
+      PSafetyMode mode = PSafeReference ///< Locking mode for the object
     );
 
     /**Create a new pointer to a PSafeObject.
@@ -393,9 +498,9 @@ class PSafePtrBase : public PObject
        idx is beyond the size of the collection, the pointer is NULL.
      */
     PSafePtrBase(
-      PSafeCollection & safeCollection, /// Collection pointer will enumerate
-      PSafetyMode mode,                 /// Locking mode for the object
-      PINDEX idx                        /// Index into collection to point to
+      const PSafeCollection & safeCollection, ///< Collection pointer will enumerate
+      PSafetyMode mode,                       ///< Locking mode for the object
+      PINDEX idx                              ///< Index into collection to point to
     );
 
     /**Create a new pointer to a PSafeObject.
@@ -406,9 +511,9 @@ class PSafePtrBase : public PObject
        otherwise the pointer is NULL.
      */
     PSafePtrBase(
-      PSafeCollection & safeCollection, /// Collection pointer will enumerate
-      PSafetyMode mode,                 /// Locking mode for the object
-      PSafeObject * obj                 /// Inital object in collection to point to
+      const PSafeCollection & safeCollection, ///< Collection pointer will enumerate
+      PSafetyMode mode,                       ///< Locking mode for the object
+      PSafeObject * obj                       ///< Inital object in collection to point to
     );
 
     /**Copy the pointer to the PSafeObject.
@@ -417,7 +522,7 @@ class PSafePtrBase : public PObject
        the PSafeObject as well.
       */
     PSafePtrBase(
-      const PSafePtrBase & enumerator   /// Pointer to copy
+      const PSafePtrBase & enumerator   ///< Pointer to copy
     );
 
   public:
@@ -433,24 +538,37 @@ class PSafePtrBase : public PObject
        two PSafePtrBase instances are pointing to the same instance.
       */
     Comparison Compare(
-      const PObject & obj   /// Other instance to compare against
+      const PObject & obj   ///< Other instance to compare against
     ) const;
   //@}
 
   /**@name Operations */
   //@{
-  public:
+    /**Return TRUE if pointer is NULL.
+      */
+    bool operator!() const { return currentObject == NULL; }
+
+    /**Get the locking mode used by this pointer.
+      */
+    PSafetyMode GetSafetyMode() const { return lockMode; }
+
     /**Change the locking mode used by this pointer.
       */
     BOOL SetSafetyMode(
-      PSafetyMode mode  /// New locking mode
+      PSafetyMode mode  ///< New locking mode
     );
 
-  protected:
-    void Assign(const PSafePtrBase & enumerator);
-    void Assign(PSafeCollection & safeCollection);
+    /**Get the associated collection this pointer may be contained in.
+      */
+    const PSafeCollection * GetCollection() const { return collection; }
+  //@}
+
+    void Assign(const PSafePtrBase & ptr);
+    void Assign(const PSafeCollection & safeCollection);
     void Assign(PSafeObject * obj);
     void Assign(PINDEX idx);
+
+  protected:
     void Next();
     void Previous();
 
@@ -465,12 +583,11 @@ class PSafePtrBase : public PObject
       NoDereference
     };
     void ExitSafetyMode(ExitSafetyModeOption ref);
-  //@}
 
   protected:
-    PSafeCollection * collection;
-    PSafeObject     * currentObject;
-    PSafetyMode       lockMode;
+    const PSafeCollection * collection;
+    PSafeObject           * currentObject;
+    PSafetyMode             lockMode;
 };
 
 
@@ -509,8 +626,8 @@ template <class T> class PSafePtr : public PSafePtrBase
        and -- operators will not work.
      */
     PSafePtr(
-      T * obj = NULL,                   /// Physical object to point to.
-      PSafetyMode mode = PSafeReference /// Locking mode for the object
+      T * obj = NULL,                   ///< Physical object to point to.
+      PSafetyMode mode = PSafeReference ///< Locking mode for the object
     ) : PSafePtrBase(obj, mode) { }
 
     /**Create a new pointer to a PSafeObject.
@@ -521,9 +638,9 @@ template <class T> class PSafePtr : public PSafePtrBase
        idx is beyond the size of the collection, the pointer is NULL.
      */
     PSafePtr(
-      PSafeCollection & safeCollection,   /// Collection pointer will enumerate
-      PSafetyMode mode = PSafeReadWrite,  /// Locking mode for the object
-      PINDEX idx = 0                      /// Index into collection to point to
+      const PSafeCollection & safeCollection, ///< Collection pointer will enumerate
+      PSafetyMode mode = PSafeReadWrite,      ///< Locking mode for the object
+      PINDEX idx = 0                          ///< Index into collection to point to
     ) : PSafePtrBase(safeCollection, mode, idx) { }
 
     /**Create a new pointer to a PSafeObject.
@@ -534,9 +651,9 @@ template <class T> class PSafePtr : public PSafePtrBase
        otherwise the pointer is NULL.
      */
     PSafePtr(
-      PSafeCollection & safeCollection, /// Collection pointer will enumerate
-      PSafetyMode mode,                 /// Locking mode for the object
-      PSafeObject * obj                 /// Inital object in collection to point to
+      const PSafeCollection & safeCollection, ///< Collection pointer will enumerate
+      PSafetyMode mode,                       ///< Locking mode for the object
+      PSafeObject * obj                       ///< Inital object in collection to point to
     ) : PSafePtrBase(safeCollection, mode, obj) { }
 
     /**Copy the pointer to the PSafeObject.
@@ -545,17 +662,17 @@ template <class T> class PSafePtr : public PSafePtrBase
        the PSafeObject as well.
       */
     PSafePtr(
-      const PSafePtr & enumerator   /// Pointer to copy
-    ) : PSafePtrBase(enumerator) { }
+      const PSafePtr & ptr   ///< Pointer to copy
+    ) : PSafePtrBase(ptr) { }
 
     /**Copy the pointer to the PSafeObject.
        This will create a copy of the pointer with the same locking mode and
        lock on the PSafeObject. It will also increment the reference count on
        the PSafeObject as well.
       */
-    PSafePtr & operator=(const PSafePtr & enumerator)
+    PSafePtr & operator=(const PSafePtr & ptr)
       {
-        Assign(enumerator);
+        Assign(ptr);
         return *this;
       }
 
@@ -563,7 +680,7 @@ template <class T> class PSafePtr : public PSafePtrBase
        This will create a read/write locked reference to teh first element in
        the collection.
       */
-    PSafePtr & operator=(PSafeCollection & safeCollection)
+    PSafePtr & operator=(const PSafeCollection & safeCollection)
       {
         Assign(safeCollection);
         return *this;
@@ -662,10 +779,40 @@ template <class T> class PSafePtr : public PSafePtrBase
         return (T *)currentObject;
       }
   //@}
+
+  /**Cast the pointer to a different type. The pointer being cast to MUST
+     be a derived class or NULL is returned.
+    */
+      /*
+  template <class Base>
+  static PSafePtr<T> DownCast(const PSafePtr<Base> & oldPtr)
+  {
+    PSafePtr<T> newPtr;
+    Base * realPtr = oldPtr;
+    if (realPtr != NULL && PIsDescendant(realPtr, T))
+      newPtr.Assign(oldPtr);
+    return newPtr;
+  }
+  */
 };
 
 
-/** This class defines a thread-safe list of objects.
+/**Cast the pointer to a different type. The pointer being cast to MUST
+    be a derived class or NULL is returned.
+  */
+template <class Base, class Derived>
+PSafePtr<Derived> PSafePtrCast(const PSafePtr<Base> & oldPtr)
+{
+//  return PSafePtr<Derived>::DownCast<Base>(oldPtr);
+    PSafePtr<Derived> newPtr;
+    Base * realPtr = oldPtr;
+    if (realPtr != NULL && PIsDescendant(realPtr, Derived))
+      newPtr.Assign(oldPtr);
+    return newPtr;
+}
+
+
+/** This class defines a thread-safe collection of objects.
 
   This is part of a set of classes to solve the general problem of a
   collection (eg a PList or PDictionary) of objects that needs to be a made
@@ -675,15 +822,15 @@ template <class T> class PSafePtr : public PSafePtrBase
   See the PSafeObject class for more details. Especially in regard to
   enumeration of collections.
  */
-template <class Coll, class Base> class PSafeList : public PSafeCollection
+template <class Coll, class Base> class PSafeColl : public PSafeCollection
 {
-    PCLASSINFO(PSafeList, PSafeCollection);
+    PCLASSINFO(PSafeColl, PSafeCollection);
   public:
   /**@name Construction */
   //@{
     /**Create a safe list collection wrapper around the real collection.
       */
-    PSafeList()
+    PSafeColl()
       : PSafeCollection(new Coll)
       { }
   //@}
@@ -694,10 +841,14 @@ template <class Coll, class Base> class PSafeList : public PSafeCollection
        This uses the PCollection::Append() function to add the object to the
        collection, with full mutual exclusion locking on the collection.
       */
-    virtual PINDEX Append(
-      Base * obj        /// Object to add to safe collection.
+    virtual PSafePtr<Base> Append(
+      Base * obj,       ///< Object to add to safe collection.
+      PSafetyMode mode = PSafeReference
     ) {
-        return PSafeCollection::SafeAppend(obj);
+        PWaitAndSignal mutex(collectionMutex);
+        if (!obj->SafeReference())
+          return NULL;
+        return PSafePtr<Base>(*this, mode, collection->Append(obj));
       }
 
     /**Remove an object to the collection.
@@ -709,7 +860,7 @@ template <class Coll, class Base> class PSafeList : public PSafeCollection
        is maintained.
       */
     virtual BOOL Remove(
-      Base * obj          /// Object to remove from safe collection
+      Base * obj          ///< Object to remove from safe collection
     ) {
         return SafeRemove(obj);
       }
@@ -722,10 +873,10 @@ template <class Coll, class Base> class PSafeList : public PSafeCollection
        As for Append() full mutual exclusion locking on the collection itself
        is maintained.
       */
-    virtual Base * RemoveAt(
-      PINDEX idx     /// Index to remove
+    virtual BOOL RemoveAt(
+      PINDEX idx     ///< Index to remove
     ) {
-        return (Base *)SafeRemoveAt(idx);
+        return SafeRemoveAt(idx);
       }
 
     /**Get the instance in the collection of the index.
@@ -733,7 +884,7 @@ template <class Coll, class Base> class PSafeList : public PSafeCollection
        PSafeObject and lock to the object in the mode specified. The lock
        will remain until the PSafePtr goes out of scope.
       */
-    virtual PSafePtr<Base> GetWithLock(
+    virtual PSafePtr<Base> GetAt(
       PINDEX idx,
       PSafetyMode mode = PSafeReadWrite
     ) {
@@ -749,13 +900,43 @@ template <class Coll, class Base> class PSafeList : public PSafeCollection
       const Base & value,
       PSafetyMode mode = PSafeReadWrite
     ) {
-        PWaitAndSignal mutex(collectionMutex);
-        return PSafePtr<Base>(*this, mode, collection->GetValuesIndex(value));
+        collectionMutex.Wait();
+        PSafePtr<Base> ptr(*this, PSafeReference, collection->GetValuesIndex(value));
+        collectionMutex.Signal();
+        ptr.SetSafetyMode(mode);
+        return ptr;
       }
   //@}
 };
 
 
+/** This class defines a thread-safe array of objects.
+  See the PSafeObject class for more details. Especially in regard to
+  enumeration of collections.
+ */
+template <class Base> class PSafeArray : public PSafeColl<PArray<Base>, Base>
+{
+};
+
+
+/** This class defines a thread-safe list of objects.
+  See the PSafeObject class for more details. Especially in regard to
+  enumeration of collections.
+ */
+template <class Base> class PSafeList : public PSafeColl<PList<Base>, Base>
+{
+};
+
+
+/** This class defines a thread-safe sorted array of objects.
+  See the PSafeObject class for more details. Especially in regard to
+  enumeration of collections.
+ */
+template <class Base> class PSafeSortedList : public PSafeColl<PSortedList<Base>, Base>
+{
+};
+
+
 /** This class defines a thread-safe dictionary of objects.
 
   This is part of a set of classes to solve the general problem of a
@@ -766,15 +947,15 @@ template <class Coll, class Base> class PSafeList : public PSafeCollection
   See the PSafeObject class for more details. Especially in regard to
   enumeration of collections.
  */
-template <class Coll, class Key, class Base> class PSafeDictionary : public PSafeCollection
+template <class Coll, class Key, class Base> class PSafeDictionaryBase : public PSafeCollection
 {
-    PCLASSINFO(PSafeDictionary, PSafeCollection);
+    PCLASSINFO(PSafeDictionaryBase, PSafeCollection);
   public:
   /**@name Construction */
   //@{
     /**Create a safe dictionary wrapper around the real collection.
       */
-    PSafeDictionary()
+    PSafeDictionaryBase()
       : PSafeCollection(new Coll) { }
   //@}
 
@@ -788,7 +969,8 @@ template <class Coll, class Key, class Base> class PSafeDictionary : public PSaf
       {
         collectionMutex.Wait();
         SafeRemove(((Coll *)collection)->GetAt(key));
-        ((Coll *)collection)->SetAt(key, obj);
+        if (obj->SafeReference())
+          ((Coll *)collection)->SetAt(key, obj);
         collectionMutex.Signal();
       }
 
@@ -801,18 +983,27 @@ template <class Coll, class Key, class Base> class PSafeDictionary : public PSaf
        is maintained.
       */
     virtual BOOL RemoveAt(
-      const Key & key   /// Key to fund object to delete
+      const Key & key   ///< Key to fund object to delete
     ) {
         PWaitAndSignal mutex(collectionMutex);
         return SafeRemove(((Coll *)collection)->GetAt(key));
       }
 
+    /**Determine of the dictionary contains an entry for the key.
+      */
+    virtual BOOL Contains(
+      const Key & key
+    ) {
+        PWaitAndSignal lock(collectionMutex);
+        return ((Coll *)collection)->Contains(key);
+      }
+
     /**Get the instance in the collection of the index.
        The returned safe pointer will increment the reference count on the
        PSafeObject and lock to the object in the mode specified. The lock
        will remain until the PSafePtr goes out of scope.
       */
-    virtual PSafePtr<Base> GetWithLock(
+    virtual PSafePtr<Base> GetAt(
       PINDEX idx,
       PSafetyMode mode = PSafeReadWrite
     ) {
@@ -828,13 +1019,25 @@ template <class Coll, class Key, class Base> class PSafeDictionary : public PSaf
       const Key & key,
       PSafetyMode mode = PSafeReadWrite
     ) {
-        PWaitAndSignal mutex(collectionMutex);
-        return PSafePtr<Base>(*this, mode, ((Coll *)collection)->GetAt(key));
+        collectionMutex.Wait();
+        PSafePtr<Base> ptr(*this, PSafeReference, ((Coll *)collection)->GetAt(key));
+        collectionMutex.Signal();
+        ptr.SetSafetyMode(mode);
+        return ptr;
       }
   //@}
 };
 
 
+/** This class defines a thread-safe array of objects.
+  See the PSafeObject class for more details. Especially in regard to
+  enumeration of collections.
+ */
+template <class Key, class Base> class PSafeDictionary : public PSafeDictionaryBase<PDictionary<Key, Base>, Key, Base>
+{
+};
+
+
 #endif // _SAFE_COLLECTION_H
 
 
